using System;
using System.Drawing;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using System.Collections;

namespace PhysicsEngine 
{
    public class StackingScene : RigidBodyList
    {
        ArrayList bodies;
        
        public StackingScene(Device device)
        {
            RigidBodyType platformType = new RigidBodyType("PlatformType", @"..\..\models\platform.x", device);
            RigidBodyType boxType = new RigidBodyType("BoxType", @"..\..\models\box.x", device);
            RigidBodyType smallboxType = new RigidBodyType("BoxType", @"..\..\models\small_box.x", device);
            RigidBodyType geoType = new RigidBodyType("GeosphereType", @"..\..\models\geosphere.x", device);
            RigidBodyType tetType = new RigidBodyType("TetrahedronType", @"..\..\models\tetrahedron.x", device);
            bodies = new ArrayList();

            for (int i = 0; i < 7; i++)
            {
                bodies.Add(new RigidBody(
                    "Box" + (i+1),                     //name
                    smallboxType,                    //type
                    new Vector3(0,1+3*i,0),         //pos
                    new Vector3(0,0,0),         //vel
                    new Quaternion(0,0,0,0),    //orientation
                    new Vector3(0,0,0),         //rotation
                    .2f,                       //bounciness
                    .5f,                           //friction
                    1.0f                        //density
                    ));
            }
            
            bodies.Add(new RigidBody(
                "Box8",                     //name
                smallboxType,                    //type
                new Vector3(-40,4,0),         //pos
                new Vector3(10,0,0),         //vel
                new Quaternion(0,0,0,0),    //orientation
                new Vector3(0,0,0),         //rotation
                .2f,                       //bounciness
                .5f,                           //friction
                float.PositiveInfinity                        //density
                ));

            bodies.Add(new RigidBody(
                "Platform1",                  //name
                platformType,                    //type
                new Vector3(0,-5,0),         //pos
                new Vector3(0,0,0),        //vel
                new Quaternion(0,0,0,0),    //orientation
                new Vector3(0,0,0),         // rotation
                .2f,                       //bounciness
                .5f,                           //friction
                float.PositiveInfinity                        //density
                ));
        }

        public int GetNumBodies() { return bodies.Count; }
        public RigidBody GetBody(int i) {  return bodies[i] as RigidBody;}
    }

    public class TowerScene : RigidBodyList
    {
        ArrayList bodies;
        
        public TowerScene(Device device)
        {
            RigidBodyType platformType = new RigidBodyType("PlatformType", @"..\..\models\platform.x", device);
            RigidBodyType smallboxType = new RigidBodyType("BoxType", @"..\..\models\small_box.x", device);
            bodies = new ArrayList();

            bodies.Add(new RigidBody(
                "Platform1",                     //name
                platformType,                    //type
                new Vector3(0,-3.2f,0),         //pos
                new Vector3(0,0,0),         //vel
                new Quaternion(0,0,0,0),    //orientation
                new Vector3(0,0,0),         //rotation
                .5f,                       //bounciness
                .5f,                           //friction
                float.PositiveInfinity      //density
                ));

            for (int i = 0; i < 5; i++)
            {
                bodies.Add(new RigidBody(
                    "Box" + (i+2),                     //name
                    smallboxType,                    //type
                    new Vector3(-5+.5f*i,-2+2.5f*i,0),         //pos
                    new Vector3(0,0,0),         //vel
                    new Quaternion(0,0,0,0),    //orientation
                    new Vector3(0,0,0),         //rotation
                    .5f,                       //bounciness
                    .5f,                           //friction
                    1.0f                        //density
                    ));
            }
        }

        public int GetNumBodies() { return bodies.Count; }
        public RigidBody GetBody(int i) {  return bodies[i] as RigidBody;}
    }

    public class SlidingScene : RigidBodyList
    {
        ArrayList bodies;
        
        public SlidingScene(Device device)
        {
            RigidBodyType platformType = new RigidBodyType("PlatformType", @"..\..\models\platform.x", device);
            RigidBodyType boxType = new RigidBodyType("BoxType", @"..\..\models\box.x", device);
            RigidBodyType smallboxType = new RigidBodyType("BoxType", @"..\..\models\small_box.x", device);
            RigidBodyType geoType = new RigidBodyType("GeosphereType", @"..\..\models\geosphere.x", device);
            RigidBodyType tetType = new RigidBodyType("TetrahedronType", @"..\..\models\tetrahedron.x", device);
            bodies = new ArrayList();

            Matrix rot = Matrix.RotationAxis(new Vector3(0,0,1), .3f);
            
            bodies.Add(new RigidBody(
                "Box2",                     //name
                smallboxType,                    //type
                Vector3.TransformCoordinate(new Vector3(8,0,0), rot),         //pos
                new Vector3(0,0,0),         //vel
                Quaternion.RotationMatrix(rot),    //orientation
                new Vector3(0,0,0),         // rotation
                1,                       //bounciness
                .5f,                           //friction
                1.0f                        //density
                ));

            bodies.Add(new RigidBody(
                "Platform1",                  //name
                platformType,                    //type
                Vector3.TransformCoordinate(new Vector3(0,-1.1f,0), rot),         //pos
                new Vector3(0,0,0),        //vel
                Quaternion.RotationMatrix(rot),    //orientation
                new Vector3(0,0,0),         //rotation
                1,                       //bounciness
                .5f,                           //friction
                float.PositiveInfinity                        //density
                ));
        }

        public int GetNumBodies() { return bodies.Count; }
        public RigidBody GetBody(int i) {  return bodies[i] as RigidBody;}
    }

    public class LaunchScene : RigidBodyList
    {
        ArrayList bodies;
        
        public LaunchScene(Device device)
        {
            RigidBodyType platformType = new RigidBodyType("PlatformType", @"..\..\models\platform.x", device);
            RigidBodyType plankType = new RigidBodyType("PlankType", @"..\..\models\plank.x", device);
            RigidBodyType boxType = new RigidBodyType("BoxType", @"..\..\models\box.x", device);
            RigidBodyType smallboxType = new RigidBodyType("SmallBoxType", @"..\..\models\small_box.x", device);
            RigidBodyType ballType = new RigidBodyType("BallType", @"..\..\models\ball.x", device);
            bodies = new ArrayList();

            for (int i = 0; i < 4; i++)
            {
                bodies.Add(new RigidBody(
                    "Box" + (i+1),                     //name
                    smallboxType,                    //type
                    new Vector3(5,0+3f*i,0),         //pos
                    new Vector3(0,0,0),         //vel
                    new Quaternion(0,0,0,0),    //orientation
                    new Vector3(0,0,0),         //rotation
                    .2f,                       //bounciness
                    .5f,                           //friction
                    1.0f                        //density
                    ));
            }

            bodies.Add(new RigidBody(
                "Box5",                     //name
                boxType,                    //type
                new Vector3(-2,10,0),         //pos
                new Vector3(0,0,0),         //vel
                new Quaternion(0,0,0,0),    //orientation
                new Vector3(0,0,0),         //rotation
                .2f,                       //bounciness
                .5f,                           //friction
                1.0f                        //density
                ));

            bodies.Add(new RigidBody(
                "Plank1",                  //name
                plankType,                    //type
                new Vector3(2f,-3f,0),         //pos
                new Vector3(0,0,0),        //vel
                new Quaternion(0,0,0,0),    //orientation
                new Vector3(0,0,0),         // rotation
                .2f,                       //bounciness
                .5f,                           //friction
                1.0f                        //density
                ));

            bodies.Add(new RigidBody(
                "Platform1",                  //name
                boxType,                    //type
                new Vector3(5f,-7,0),         //pos
                new Vector3(0,0,0),        //vel
                new Quaternion(0,0,0,0),    //orientation
                new Vector3(0,0,0),         // rotation
                .2f,                       //bounciness
                .5f,                           //friction
                float.PositiveInfinity                        //density
                ));
        }

        public int GetNumBodies() { return bodies.Count; }
        public RigidBody GetBody(int i) {  return bodies[i] as RigidBody;}
    }

    public class SeesawScene : RigidBodyList
    {
        ArrayList bodies;
        
        public SeesawScene(Device device)
        {
            RigidBodyType thickplankType = new RigidBodyType("PlankType", @"..\..\models\thickplank.x", device);
            RigidBodyType platformType = new RigidBodyType("PlatformType", @"..\..\models\platform.x", device);
            RigidBodyType smallboxType = new RigidBodyType("BoxType", @"..\..\models\small_box.x", device);
            RigidBodyType smallerboxType = new RigidBodyType("BoxType", @"..\..\models\smaller_box.x", device);
            RigidBodyType rollerType = new RigidBodyType("GeosphereType", @"..\..\models\roller.x", device);
            bodies = new ArrayList();

            bodies.Add(new RigidBody(
                "Box1",                     //name
                smallboxType,                    //type
                new Vector3(3.5f,2.8f,0),         //pos
                new Vector3(0,0,0),         //vel
                new Quaternion(0,0,0,0),    //orientation
                new Vector3(0,0,0),         // rotation
                .2f,                       //bounciness
                .1f,                           //friction
                1.0f                        //density
                ));
 
            bodies.Add(new RigidBody(
                "Box2",                     //name
                smallerboxType,                    //type
                new Vector3(-4f,2.3f,0),         //pos
                new Vector3(0,0,0),         //vel
                new Quaternion(0,0,0,0),    //orientation
                new Vector3(0,0,0),         // rotation
                .2f,                       //bounciness
                .8f,                           //friction
                1.0f                        //density
                ));

            bodies.Add(new RigidBody(
                "Plank1",                  //name
                thickplankType,                    //type
                new Vector3(0,1.5f,0),         //pos
                new Vector3(0,0,0),        //vel
                new Quaternion(0,0,0,0),    //orientation
                new Vector3(0,0,0),         //rotation
                .2f,                       //bounciness
                .1f,                           //friction
                1.0f                        //density
                ));

            bodies.Add(new RigidBody(
                "Roller1",                  //name
                rollerType,                    //type
                new Vector3(0,.6f,0),         //pos
                new Vector3(0,0,0),        //vel
                new Quaternion(0,0,0,0),    //orientation
                new Vector3(0,0,0),         //rotation
                .2f,                       //bounciness
                1,                           //friction
                float.PositiveInfinity                        //density
                ));

            bodies.Add(new RigidBody(
                "Platform1",                  //name
                platformType,                    //type
                new Vector3(0,-.1f,0),         //pos
                new Vector3(0,0,0),        //vel
                new Quaternion(0,0,0,0),    //orientation
                new Vector3(0,0,0),         //rotation
                .2f,                       //bounciness
                .2f,                           //friction
                float.PositiveInfinity                        //density
                ));
        }

        public int GetNumBodies() { return bodies.Count; }
        public RigidBody GetBody(int i) {  return bodies[i] as RigidBody;}
    }

    
    public class SkullScene : RigidBodyList
    {
        ArrayList bodies;
        
        public SkullScene(Device device)
        {
            RigidBodyType platformType = new RigidBodyType("PlatformType", @"..\..\models\platform.x", device);
            RigidBodyType skullType = new RigidBodyType("GeosphereType", @"..\..\models\skull.x", device);
            RigidBodyType brainType = new RigidBodyType("GeosphereType", @"..\..\models\brain.x", device);
            bodies = new ArrayList();

            bodies.Add(new RigidBody(
                "Skull1",                     //name
                skullType,                    //type
                new Vector3(0,10,-2),         //pos
                new Vector3(0,0,0),         //vel
                new Quaternion(0,0,0,0),    //orientation
                new Vector3(0,0,0),         // rotation
                .5f,                       //bounciness
                .2f,                           //friction
                1.0f                        //density
                ));
 
            bodies.Add(new RigidBody(
                "Brain1",                     //name
                brainType,                    //type
                new Vector3(0,13,-2),         //pos
                new Vector3(0,0,0),         //vel
                new Quaternion(0,0,0,0),    //orientation
                new Vector3(0,0,0),         // rotation
                .5f,                       //bounciness
                .2f,                           //friction
                1.0f                        //density
                ));

            bodies.Add(new RigidBody(
                "Platform1",                  //name
                platformType,                    //type
                new Vector3(0,-.1f,0),         //pos
                new Vector3(0,0,0),        //vel
                new Quaternion(0,0,0,0),    //orientation
                new Vector3(0,0,0),         //rotation
                .5f,                       //bounciness
                .2f,                           //friction
                float.PositiveInfinity                        //density
                ));
        }

        public int GetNumBodies() { return bodies.Count; }
        public RigidBody GetBody(int i) {  return bodies[i] as RigidBody;}
    }

    public class PhysicsEngine : GraphicsApplication
    {
        private const string stateFileName = @"C:\Documents and Settings\Evan\My Documents\classes\cs448a physics\project\PhysicsEngine\initial_state.xml";
        
        private WorldState worldState;
        private Device device;
        private Timer timer;

        private float lastTime;
        private float time_step_size;

        public PhysicsEngine()
        {
            this.Text = "PhysicsEngine";
            this.KeyDown += new KeyEventHandler(this.OnPrivateKeyDown);
            this.KeyUp += new KeyEventHandler(this.OnPrivateKeyUp);
            timer = new Timer();
            lastTime = 0;
        }

        private void OnPrivateKeyDown(object sender, KeyEventArgs e)
        {
            switch (e.KeyCode)
            {
                case Keys.Up:
                {
                    break;
                }
                case Keys.Down:
                {
                    break;
                }
                case Keys.Left:
                {
                    break;
                }
                case Keys.Right:
                {
                    break;
                }
                case Keys.Space:
                {
                    timer.Running = !timer.Running;
                    break;
                }
                case Keys.Escape:
                {
                    Form.ActiveForm.Close();
                    break;
                }
                case Keys.D1:
                {
                    Console.Out.WriteLine("Loading STACKING scene...");
                    worldState.BodyList = new StackingScene(device);
                    Console.Out.WriteLine("done");
                    break;
                }
                case Keys.D2:
                {
                    Console.Out.WriteLine("Loading TOWER scene...");
                    worldState.BodyList = new TowerScene(device);
                    Console.Out.WriteLine("done");
                    break;
                }
                case Keys.D3:
                {
                    Console.Out.WriteLine("Loading SLIDING scene...");
                    worldState.BodyList = new SlidingScene(device);
                    Console.Out.WriteLine("done");
                    break;
                }
                case Keys.D4:
                {
                    Console.Out.WriteLine("Loading LAUNCH scene...");
                    worldState.BodyList = new LaunchScene(device);
                    Console.Out.WriteLine("done");
                    break;
                }
                case Keys.D5:
                {
                    Console.Out.WriteLine("Loading SEESAW scene...");
                    worldState.BodyList = new SeesawScene(device);
                    Console.Out.WriteLine("done");
                    break;
                }
                case Keys.D6:
                {
                    Console.Out.WriteLine("Loading SKULL scene...");
                    worldState.BodyList = new SkullScene(device);
                    Console.Out.WriteLine("done");
                    break;
                }
                case Keys.N:
                {
                    if (worldState.NumCollisionSteps > 0) 
                    {
                        worldState.NumCollisionSteps--;
                        Console.Out.WriteLine("Decreasing number of collision steps to {0}", worldState.NumCollisionSteps);
                    }
                    break;
                }
                case Keys.M:
                {
                    worldState.NumCollisionSteps++;
                    Console.Out.WriteLine("Increasing number of collision steps to {0}", worldState.NumCollisionSteps);
                    break;
                }
                case Keys.J:
                {
                    if (worldState.NumContactSteps > 0) 
                    {
                        worldState.NumContactSteps--;
                        Console.Out.WriteLine("Decreasing number of contact steps to {0}", worldState.NumContactSteps);
                    }
                    break;
                }
                case Keys.K:
                {
                    worldState.NumContactSteps++;
                    Console.Out.WriteLine("Increasing number of contact steps to {0}", worldState.NumContactSteps);
                    break;
                }
                case Keys.Y:
                {
                    worldState.EdgeIntersection = !worldState.EdgeIntersection;
                    Console.Out.WriteLine("Turning edge-edge intersection testing {0}", worldState.EdgeIntersection ? "ON" : "OFF");
                    break;
                }
                case Keys.P:
                {
                    worldState.ShockPropagation = !worldState.ShockPropagation;
                    Console.Out.WriteLine("Turning shock propagation {0}", worldState.ShockPropagation ? "ON" : "OFF");
                    break;
                }
                case Keys.O:
                {
                    worldState.CollisionsBeforeVelocityUpdate = !worldState.CollisionsBeforeVelocityUpdate;
                    Console.Out.WriteLine("Now doing collisions {0} velocity update", worldState.CollisionsBeforeVelocityUpdate ? "BEFORE" : "AFTER");
                    break;
                }
            }
        }
        private void OnPrivateKeyUp(object sender, KeyEventArgs e)
        {
            switch (e.KeyCode)
            {
                case Keys.Up:
                case Keys.Down:
                    break;
                case Keys.Left:
                case Keys.Right:
                {
                    break;
                }
            }
        }

        public override bool InitializeGraphics()
        {
            PresentParameters pres = new PresentParameters(); 
            pres.Windowed = true; 
            pres.SwapEffect = SwapEffect.Discard;
            pres.EnableAutoDepthStencil = true; 
            pres.AutoDepthStencilFormat = DepthFormat.D16;

            device = new Device(0, DeviceType.Hardware, this, 
                CreateFlags.SoftwareVertexProcessing, pres); 

            // XXX is this function ever called more than once?? 
            // shouldn't reload entire state if so
            worldState = new WorldState(new StackingScene(device), 2, 2, true, true, true);
            //time_step_size = 0.005f;
            time_step_size = 0.0083f;

            return true;   
        }

        public override void DisposeGraphics()
        {
        }

        private void SetupLights()
        {
            Material mtrl = new Material();
            mtrl.Diffuse = Color.White;
            device.Material = mtrl;
            
            device.RenderState.Lighting = true;

            device.Lights[0].Type = LightType.Directional;
            device.Lights[0].Direction = new Vector3(0.3f, -0.5f, 0.2f);
            device.Lights[0].Diffuse = Color.White;
            device.Lights[0].Commit();
            device.Lights[0].Enabled = true;
        }

        public override void Render()
        {
            //update the world
            if (timer.Running)
            {
                float curTime = timer.CurrentTime;
                float dt = curTime - lastTime;
                int max_steps = (int)(1.0f / (60.0f * time_step_size)); // never do more than 1/60th of a second between frames
                int num_steps = (int)Math.Ceiling(dt / time_step_size);
                num_steps = Math.Min(num_steps, max_steps);
                num_steps = Math.Max(num_steps, 1); // always do at least 1 step
                for (int i = 0; i < num_steps; i++)
                    worldState.Update(time_step_size);
                lastTime = curTime;
            }
            //Clear the backbuffer to a Bisque color 
            device.Clear(ClearFlags.Target | ClearFlags.ZBuffer, Color.Black, 1.0f, 0);
            
            device.BeginScene();

            SetupLights();

            // Setup the world, view, and projection matrices
            device.Transform.View = Matrix.LookAtLH( new Vector3( 0.0f, 10.0f, -16.0f), new Vector3(0,0,0), new Vector3( 0.0f, 1.0f, 0.0f));   
            device.Transform.Projection = Matrix.PerspectiveFovLH( (float)Math.PI / 4, (float)Width/Height, 1.0f, 100.0f);

            // Render the world
            worldState.Render(device);

            device.EndScene();

            device.Present();
        }
    }
}