using System;

namespace PhysicsEngine
{
	/// <summary>
	/// Summary description for Timer.
	/// </summary>
	public class Timer
	{
        private int startTickCount;
        private int stopTickCount;
        
        private bool running;

        public Timer()
		{
            Reset();
            running = false;
		}

        public void Reset()
        {
            startTickCount = stopTickCount = Environment.TickCount;
        }

        public bool Running
        {
            get { return running; }
            set
            {
                if (running == value) return;
                if (value == true) // start the timer
                    startTickCount = Environment.TickCount - (stopTickCount - startTickCount);
                else               // stop the timer
                    stopTickCount = Environment.TickCount;
                running = value;
            }
        }
        
        // returns the total amount of time elapsed since the timer was started,
        // minus any gaps when it was stopped
        public float CurrentTime
        {
            get { return CurrentTickCount / 1000.0f; }
        }

        private int CurrentTickCount
        {
            get
            {
                if (running)
                    return Environment.TickCount - startTickCount;
                else
                    return stopTickCount - startTickCount;
            }
        }
	}
}
